import json

import click

from hpogrid.core.defaults import *

class ListOption(click.Option):
    def type_cast_value(self, ctx, value):
        if isinstance(value, str):
            return [i.strip() for i in value.split(',') if i.strip()]
        return value
    
class DictOption(click.Option):
    def type_cast_value(self, ctx, value):
        if isinstance(value, str):
            return json.loads(value)
        return value    
    
@click.command(name='generate')
@click.option('-l', '--lib', default=kDefaultGenerator, type=click.Choice(kGenerators, case_sensitive=False),
              help='optimization library to use', show_default=True)
@click.option('-s', '--space', required=True, 
              help='json file containing search space definition')
@click.option('-n', '--n_point', type=int, required=True, 
              help='number of search points to generate')
@click.option('-m', '--max_point', type=int, default=None, 
              help='(iDDS only) maximum number of points to generate in entire iDDS workflow')
@click.option('--metric', default=kDefaultMetric, 
              help='evaluation metric', show_default=True)
@click.option('--mode', default=kDefaultMetricMode, type=click.Choice(kMetricMode, case_sensitive=False),
              help='evaluation mode', show_default=True)
@click.option('-i', '--infile', default=None,
              help='(iDDS only) iDDS input')
@click.option('-o', '--outfile', default=None,
              help='(iDDS only) iDDS output')
@click.option('--method', default=None,
              help='(nevergrad library only) optimizer type')
def generate(**kwargs):
    """
    Generate hyperparameter search points
    """
    from hpogrid.interface.idds.steering import SteeringIDDS
    SteeringIDDS().run_generator(**kwargs)
    
@click.command(name='idds_log', short_help='Tool for retrieving iDDS logs')
@click.argument('ID', type=int)
def idds_log(**kwargs):
    """
    Tool for retrieving iDDS logs
    
    ID: JEDI task ID for the iDDS job
    """
    from hpogrid.interface.main import download_log
    kwargs['workload_id'] = kwargs.pop('id')
    download_log(**kwargs)
    
@click.command(name='sites')
@click.option('-n', '--name', default=None, 
              help='filter grid site by name')
@click.option('--gpu_sites/--all_sites', 'gpu', default=True, 
              help='show gpu sites only or all sites', show_default=True)
@click.option('--active/--all_status', default=True, 
              help='show active sites only or sites with any status', show_default=True)
@click.option('-t', '--site_type', default=kDefultGridSiteType, cls=ListOption,
              help='filter sites by type', show_default=True)
@click.option('-i', '--info', default=kDefaultGridSiteInfo, cls=ListOption, 
              help='list of site info to show, separated by commas', show_default=True)
def display_sites(**kwargs):
    """
    Show available grid sites
    """
    from hpogrid.components import GridSiteInfo
    GridSiteInfo.show(**kwargs)


import os
import argparse
import tarfile

kDefaultDisk = 'UKI-NORTHGRID-MAN-HEP_SCRATCHDISK'
kDefaultDatasetName = 'user.${RUCIO_ACCOUNT}.hpo.test.dataset.01'
kDefaultTempDir = './'

def get_rucio_parser():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('data_paths', nargs='+', help=_help_test)
    parser.add_argument('-n', '--name', help=_help_test, default=kDefaultDatasetName)
    parser.add_argument('-t', '--temp_dir', help=_help_test, default=kDefaultTempDir)
    parser.add_argument('-d','--disk', help=_help_test, default=kDefaultDisk)
    parser.add_argument('-s', '--save', action='store_true', help=_help_test, default=False)

    return parser


def upload_rucio(args):
    disk_name = args.disk
    dataset_name = args.name
    tarfile_name = args.name.replace('user.${RUCIO_ACCOUNT}.','')+'.tar.gz'
    tarfile_fullpath = os.path.join(args.temp_dir, tarfile_name)
    if not os.path.exists(tarfile_fullpath):
        make_tarfile(args.data_paths, tarfile_fullpath)

    os.system('rucio add-dataset {}'.format(dataset_name))
    os.system('rucio upload {} --name {} --rse {}'.format(tarfile_fullpath, tarfile_name, disk_name))
    os.system('rucio attach {} user.${{RUCIO_ACCOUNT}}:{}'.format(dataset_name, tarfile_name))


def make_tarfile(source, outname):
    print('INFO: Making tarfile for {}'.format(source))
    with tarfile.open(outname, "w:gz") as tar:
        for src in source:
            src_path = os.path.abspath(os.path.expanduser(src.rstrip('/')))
            tar.add(src_path, arcname=os.path.basename(src_path))

if __name__ == '__main__':
    parser = get_rucio_parser()
    args = parser.parse_args()
    upload_rucio(args)



from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

class BaseAlgorithm():
    
    def __init__(self, metric:str, mode:str, search_space:Dict):
        self.metric = metric
        self.mode = mode
        self.search_space = search_space
        self.optimizer = None
        self.algorithm = None
    
    @abstractmethod
    def _get_algorithm(self, max_concurrent:int=None, **args):
        pass
    
    @abstractmethod
    def _restore_search_points(self, algorithm, restore_points:Tuple):
        pass
    
    def create(self, max_concurrent:int=None, restore_points:Optional[Tuple]=None, **args):
        algorithm = self._get_algorithm(**args)
        # restore search points
        if restore_points:
            self._restore_search_points(algorithm, restore_points)
        # limit max concurrency
        if algorithm and max_concurrent:
            from ray.tune.suggest import ConcurrencyLimiter
            algorithm = ConcurrencyLimiter(algorithm, max_concurrent=max_concurrent)
        self.algorithm = algorithm
        
        return algorithm


import math

from .base_space import BaseSpace

class BOHBSpace(BaseSpace):

    def __init__(self, search_space=None):
        self.library = 'bohb'
        super().__init__(search_space)    
        
    def reset_space(self):
        import ConfigSpace as CS
        self.search_space = CS.ConfigurationSpace(seed=1)    

    def append(self, space_value):
        self.search_space.add_hyperparameter(space_value)

    def categorical(self, label, categories, grid_search = False):
        if grid_search != False:
            raise ValueError(f'{self.library} does not allow grid search')
        import ConfigSpace.hyperparameters as CSH
        return CSH.CategoricalHyperparameter(name=label, choices=categories)

    def uniformint(self, label, low, high):
        import ConfigSpace.hyperparameters as CSH
        return CSH.UniformIntegerHyperparameter(name=label, lower=low, upper=high)

    def uniform(self, label, low, high):
        import ConfigSpace.hyperparameters as CSH
        return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high)

    def quniform(self, label, low, high, q):
        import ConfigSpace.hyperparameters as CSH
        return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high, q=q)

    def qloguniform(self, label, low, high, q, base=math.e):
        if base != math.e:
            raise ValueError(f'{self.library} only allows base e for qloguniform')
        import ConfigSpace.hyperparameters as CSH
        return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high, q=q, log=True)

    def loguniform(self, label, low, high, base=math.e):
        if base != math.e:
            raise ValueError(f'{self.library} only allows base e for loguniform')
        import ConfigSpace.hyperparameters as CSH
        return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high, log=True)

    def normal(self, label, mu, sigma):
        import ConfigSpace.hyperparameters as CSH
        return CSH.NormalFloatHyperparameter(name=label, mu=mu, sigma=sigma)

    def qnormal(self, label, mu, sigma, q):
        import ConfigSpace.hyperparameters as CSH
        return CSH.NormalFloatHyperparameter(name=label, mu=mu, sigma=sigma, q=q)

    def qlognormal(self, label, mu, sigma, q, base=math.e):
        import ConfigSpace.hyperparameters as CSH
        if base != math.e:
            raise ValueError(f'{self.library} only allows base e for qlognormal')
        return CSH.NormalFloatHyperparameter(name=label, mu=mu, sigma=sigma, q=q, log=True)

    def fixed(self, label, value):
        import ConfigSpace.hyperparameters as CSH
        return CSH.CategoricalHyperparameter(name=label, choices=[value])



def run():
    from hpogrid.generators import *
    from hpogrid.search_space import BaseSpace
    
    search_space = BaseSpace.test_space_base_e
    search_space_base_e = BaseSpace.test_space_base_e
    metric = "loss"
    mode = "min"
    